#Requires -Version 5.1
#Requires -Modules @{ ModuleName="Microsoft.Graph.Applications"; RequiredVersion="2.24.0" }
#Requires -Modules @{ ModuleName="Microsoft.Graph.Identity.SignIns"; RequiredVersion="2.24.0" }

Param(
    [Parameter(Mandatory = $true)] 
    [ValidateNotNull()]
    [string] 
    $TenantId,

    [Parameter(Mandatory = $true)] 
    [ValidateNotNull()]
    [string] 
    $AppName,

    [Parameter(Mandatory = $false)] 
    [ValidateNotNull()]
    [int]
    $ApiWebPort = 8443,

    [Parameter(Mandatory = $false)] 
    [ValidateNotNull()]
    [ValidateRange(1, 24)]
    [int] 
    $ClientSecretExpiresInMonths = 12
)

<######################################################################
Grant application consent.
######################################################################>
function Grant-Consent {
    param (
        $RequiredResourcesAccess, 
        $ServicePrincipal
    )

    Try {
        ForEach ($resourceAppAccess in $requiredResourcesAccess) {
            $resourceApp = Get-MgServicePrincipal -Filter "appId eq '$($resourceAppAccess.ResourceAppId)'"
            ForEach ($permission in $resourceAppAccess.ResourceAccess | Where-Object { $_.Type -eq "Role" }) {
                $params = @{
                    "PrincipalId" = $ServicePrincipal.Id
                    "ResourceId" = $ResourceApp.Id
                    "AppRoleId" = $permission.Id
                }
                
                New-MgServicePrincipalAppRoleAssignment -ServicePrincipalId $ServicePrincipal.Id -BodyParameter $params | Out-Null
            }
        }

        ForEach ($resourceAppAccess in $requiredResourcesAccess) {
            $delegatedPermissions = @()
            $resourceApp = Get-MgServicePrincipal -Filter "appId eq '$($resourceAppAccess.ResourceAppId)'"
            ForEach ($permission in $resourceAppAccess.ResourceAccess | Where-Object { $_.Type -eq "Scope" }) {
                $permissionObj = $resourceApp.OAuth2PermissionScopes | Where-Object { $_.Id -contains $permission.Id }
                $delegatedPermissions += $permissionObj.Value
            }

            if ($delegatedPermissions) {
                $params = @{
                    "ClientId" = $ServicePrincipal.Id
                    "ConsentType" = "AllPrincipals"
                    "ResourceId" =  $ResourceApp.Id
                    "Scope" = $delegatedPermissions -Join " "
                  }
                  
                  New-MgOauth2PermissionGrant -BodyParameter $params | Out-Null
            }
        }
    }
    Catch {
        Write-Host -f Red "Failed to grant consent."
        Exit
    }
}

<######################################################################
Add client application.
######################################################################>
function Add-ClientApplication {
    param (
        $App
    )

    Try {
        $Oauth2Permission = $App.Api.OAuth2PermissionScopes | Where-Object { $_.Value -eq 'user_impersonation' }

        $Api = @{
            PreAuthorizedApplications = @(
                @{
                    AppId = $App.AppId
                    DelegatedPermissionIds = $Oauth2Permission.Id
                }
            )
        }

        Update-MgApplication -ApplicationId $App.Id -Api $Api | Out-Null
    }
    Catch {
        Write-Host -f Red "Failed to add client application."
        Exit
    }
}

#####################################################################
# Connect to Microsoft Graph
#####################################################################
Write-Progress -Activity "Connecting to Microsoft Graph"
Connect-MgGraph -TenantId $TenantId -Scopes "Application.Read.All", "Application.ReadWrite.All", "User.Read.All", "DelegatedPermissionGrant.ReadWrite.All", "AppRoleAssignment.ReadWrite.All" | Out-Null

#####################################################################
# Check if application registration already exists
#####################################################################
if (Get-MgApplication -Filter "DisplayName eq '$($AppName)'" -ErrorAction SilentlyContinue) {
    Write-Host -f Red "Azure AD Application $AppName already exists. The Web Application has not been registered with Entra ID."
    Exit 1
}

#####################################################################
# Get Address
#####################################################################
$ApiWebAddress = [System.Net.Dns]::GetHostEntry("localhost").HostName
$Ip = [System.Net.Dns]::GetHostEntry([System.Net.Dns]::GetHostName()).AddressList |
      Where-Object { $_.AddressFamily -eq [System.Net.Sockets.AddressFamily]::InterNetwork } |
      Select-Object -First 1
$ApiWebIp = $Ip.ToString()

#####################################################################
# Create Application
#####################################################################

#####################################################################
# Manage -> API permissions -> Configured permissions
#####################################################################
$RequiredResourceAcesss = @(
    @{
        ResourceAppId = "00000003-0000-0000-c000-000000000000"
        ResourceAccess = @(
            @{
                Id = "5f8c59db-677d-491f-a6b8-5f174b11ec1d" #Group.Read.All Delegated
                Type = "Scope"
            },
            @{
                Id = "7427e0e9-2fba-42fe-b0c0-848c9e6a8182" #offline_access Delegated
                Type = "Scope"
            }
            @{
                Id = "37f7f235-527c-4136-accd-4a02d197296e" #openid Delegated
                Type = "Scope"
            },
            @{
                Id = "14dad69e-099b-42c9-810b-d002981feec1" #profile Delegated
                Type = "Scope"
            },
            @{
                Id = "e1fe6dd8-ba31-4d61-89e7-88639da4683d" #User.Read Delegated
                Type = "Scope"
            },
            @{
                Id = "df021288-bdef-4463-88db-98f22de89214" #User.Read.All Application
                Type = "Role"
            }
        )
    }
)

#####################################################################
# Manage -> App roles -> App roles
#####################################################################
$AppRoles = @( 
    @{
        AllowedMemberTypes = @("User", "Application")
        DisplayName = "administrator"
        Id = New-Guid
        IsEnabled = $true
        Description = "Kelverion Automation Portal administrator role."
        Value = "administrator"
    },
    @{
        AllowedMemberTypes = @("User")
        DisplayName = "operator"
        Id = New-Guid
        IsEnabled = $true
        Description = "Kelverion Automation Portal operator role."
        Value = "operator"
    },
    @{
        AllowedMemberTypes = @("User")
        DisplayName = "developer"
        Id = New-Guid
        IsEnabled = $true
        Description = "Kelverion Automation Portal developer role."
        Value = "developer"
    }
)

#####################################################################
# Manage -> Autentication -> Mobile and desktop applications -> Redirect URIs
# Manage -> Autentication -> iOS / macOS -> Redirect URIs
#####################################################################
$PublicClient = @{ RedirectUris = @("https://login.microsoftonline.com/common/oauth2/nativeclient", "msauth.com.kelverion.AutomationPortal://auth") } 

#####################################################################
# Manage -> Authentication -> Implicit grant and hybrid flows
# Manage -> Autentication -> Web -> Redirect URIs
#####################################################################
$Web = @{ 
    ImplicitGrantSettings = @{
        EnableAccessTokenIssuance = $true
        EnableIdTokenIssuance = $true
    }
    RedirectUris = @("https://$($ApiWebAddress):$($ApiWebPort)/oauth2-redirect.html", "https://$($ApiWebIp):$($ApiWebPort)/oauth2-redirect.html")
    HomePageUrl = "https://$($ApiWebAddress)"
} 

#####################################################################
# Manage -> Autentication -> Single-page application -> Redirect URI
#####################################################################
$Spa = @{ RedirectUris = @("https://$($ApiWebAddress)/", "https://$($ApiWebIp)/") } 

#####################################################################
# Manage -Expose an API
#####################################################################
$Api = @{
    oauth2PermissionScopes = @(
        @{
            AdminConsentDescription = "Allow the application to access $AppName on behalf of the signed-in user."
            AdminConsentDisplayName = "Access $AppName"
            UserConsentDescription  = "Allow the application to access $AppName on your behalf."
            UserConsentDisplayName  = "Access $AppName"
            Id                      = New-Guid
            IsEnabled               = $true
            Type                    = "User"
            Value                   = "user_impersonation"
        }
    )
}

Write-Progress -Activity "Creating App Registration"
$App = New-MgApplication -DisplayName $AppName `
    -AppRoles $AppRoles `
    -GroupMembershipClaims "SecurityGroup" `
    -SignInAudience "AzureADMyOrg" `
    -Web $Web `
    -Api $Api `
    -PublicClient $PublicClient `
    -Spa $Spa `
    -Tags @("KelverionAutomationPortal", "v1")

$AdminRole = $App.AppRoles.where({ $_.DisplayName -eq "administrator"})
$AdministatorPermission = @{
    ResourceAppId = $App.AppId
    ResourceAccess = @(
        @{
            Id = $AdminRole.Id
            Type = "Role"
        }
    )
}
$RequiredResourceAcesss += $AdministatorPermission

Write-Progress -Activity "Updating required resource access"
Update-MgApplication -ApplicationId $App.Id -RequiredResourceAccess $RequiredResourceAcesss | Out-Null

# Assign owner - This is a fix for New-MgApplication not assigning user as documented
Write-Progress -Activity "Assigning application owner"
$User= Invoke-MgGraphRequest -Method GET https://graph.microsoft.com/v1.0/me
$ObjectId = $User.ID
$NewOwner = @{
     "@odata.id"= "https://graph.microsoft.com/v1.0/directoryObjects/{$ObjectId}"
     }
New-MgApplicationOwnerByRef -ApplicationId $App.Id -BodyParameter $NewOwner | Out-Null

#####################################################################
# Manage -> Certificates & secrets 
#####################################################################
Write-Progress -Activity "Creating client secret"
$PasswordCred = @{
    "displayName" = "AppAccessKey"
    "endDateTime" = (Get-Date).AddMonths($ClientSecretExpiresInMonths)
}
$ClientSecret = Add-MgApplicationPassword -ApplicationId $App.Id -PasswordCredential $PasswordCred

#####################################################################
# Azure Active Directory -> Manage -> Enterprise application
#####################################################################
Write-Progress -Activity "Creating enterpise application"
$ServicePrincipalID= @{ 
    "AppId" = $App.AppId
    "Tags" = @("WindowsAzureActiveDirectoryIntegratedApp")
}
$ServicePrincipal = New-MgServicePrincipal -BodyParameter $ServicePrincipalID

#####################################################################
# Azure Active Directory -> Manage -> Enterprise application -> Manage -> Users and groups -> Add user/group
#####################################################################
Write-Progress -Activity "Creating administrator"
$AppRole = $ServicePrincipal.AppRoles | Where-Object { $_.DisplayName -eq "Administrator" }
$Owner = Get-MgApplicationOwner -ApplicationId $App.Id
New-MgUserAppRoleAssignment -UserId $Owner.Id -PrincipalId $Owner.Id -ResourceId $ServicePrincipal.Id -AppRoleID $AppRole.Id | Out-Null

#####################################################################
# Manage -> API permissions -> Configured permissions -> Grant admin consent for
#####################################################################
Write-Progress -Activity "Granting consent"
Grant-Consent -RequiredResourcesAccess $RequiredResourceAcesss -ServicePrincipal $ServicePrincipal

#####################################################################
# Manage -> Expose an Api -> Add client applications
#####################################################################
Write-Progress -Activity "Adding client application"
Add-ClientApplication -App $App

#####################################################################
# Manage -> Expose an API -> Application ID URI
#####################################################################
Write-Progress -Activity "Updating application registration"
Update-MgApplication -ApplicationId $App.Id -IdentifierUris @("api://$($App.AppId)") -IsFallbackPublicClient | Out-Null

#####################################################################
# Disonnect from Microsoft Graph
#####################################################################
Write-Progress -Activity "Disconnecting from Microsoft Graph"
Disconnect-MgGraph | Out-Null

Write-Host -ForegroundColor Yellow "Directory (tenant) ID: $($TenantId)"
Write-Host -ForegroundColor Yellow "Application (client) ID: $($App.AppId)"
Write-Host -ForegroundColor Yellow "Client Secret: $($ClientSecret.SecretText)"
Write-Host -ForegroundColor Yellow "Automation Portal Client Secret will expire on: $($ClientSecret.EndDateTime.ToString())"
Write-Host -ForegroundColor Yellow "Automation Portal API URL: $("https://$($ApiWebAddress):$($ApiWebPort)")"

Exit 0